using System;
using System.Collections;
using Utils;

namespace OPSM
{
	/// <summary>
	/// 
	/// </summary>
	public class DFSMiner : Miner
	{
		public DFSMiner(Dataset ds, DualCompare dc)
		{
			this.ds = ds;
			dualComp = dc;
		}

		override public void Mine(int support, int minLength, int maxLength, int maxMistakes, MineResults mineResult)
		{
			Itemset head = new Itemset();
			ArrayList tail = new ArrayList();

			for (int loop = 0; loop < ds.GetColumnCount(); loop++)
				tail.Add(loop);

			RecurseMining(head, tail, support, minLength, maxLength, mineResult);
		}

		void RecurseMining(Itemset head, ArrayList tail, int support, int minLength, int maxLength, MineResults mineResult)
		{
			Itemset newHead;
			ArrayList newTail;

			if (head.Count >= maxLength)
				return;

			if (head.Count + tail.Count < minLength)
				return;

			ArrayList bitArrays = new ArrayList();
			for (int loopTail = 0; loopTail < tail.Count; loopTail++)
			{
				int i = (int)tail[loopTail];

				if (head.Count == 0)
				{
					newTail = (ArrayList)tail.Clone();
					newTail.RemoveAt(loopTail);
					newHead = new Itemset();
					newHead.AddItem(i);
					RecurseMining(newHead, newTail, support, minLength, maxLength, mineResult);
				}
				else
				{
					FastSparseBitArray bitArray = null;
					
					if (dualComp != null)
					{
						ISimpleItemset dualItemset = dualComp.GetItemset(head.GetLastItem(), i);
						if (dualItemset == null)
						{
							tail.RemoveAt(loopTail);
							loopTail--;
							continue;
						}

						bitArray = dualItemset.GetTransactions();
					}
					else
					{
						bitArray = FastSparseBitArrayPool.bitArrayPool.Allocate();
						ds.BuildBitVector(head.GetLastItem(), i, bitArray);
					}

					if (head.Count > 1)
						bitArray = bitArray.And(head.GetTransactions());

					bitArray.frequency = bitArray.CountElements();
					if (bitArray.frequency >= support)
					{
						bitArrays.Add(bitArray);
					}
					else
					{
							// Don't release bit vectors from O2 matrix, or there is no
							// O2 matrix
						if ((head.Count > 1)||(dualComp == null))
							FastSparseBitArrayPool.bitArrayPool.Release(bitArray);

						tail.RemoveAt(loopTail);
						loopTail--;
					}
				}				
			}

			if (head.Count > 0)
				for (int loopTail = 0; loopTail < tail.Count; loopTail++)
				{
					int i = (int)tail[loopTail];

					newTail = (ArrayList)tail.Clone();
					newTail.RemoveAt(loopTail);
					newHead = new Itemset(head);
					newHead.AddItem(i);
					FastSparseBitArray bitArray = (FastSparseBitArray)bitArrays[loopTail];
					newHead.SetTransactions(bitArray);
					newHead.support = bitArray.frequency;
					RecurseMining(newHead, newTail, support, minLength, maxLength, mineResult);

					if (newHead.Count >= minLength)
						mineResult.Add(newHead);

					newHead.SetTransactions(null);
					if (newHead.Count > 2)
						FastSparseBitArrayPool.bitArrayPool.Release(bitArray);
				}
		}
	}	
}
